import warnings
from pathlib import Path
from glob import iglob
import os.path

from . import convert, TransmogripyWarning, FatalTransmogripyWarning


def trans_dir(glob_path, dst_root):
    root_path = glob_path[:glob_path.find('*')]
    if os.path.isdir(glob_path):
        glob_path = os.path.join(glob_path, r'**\*.pas')
    display = not glob_path.endswith('*.pas')

    blacklist = []
    files = iglob(glob_path)

    count = {'ok': 0, 'skipped': 0, 'fatal': 0, 'warnings': 0}

    for f in files:
        if f in blacklist:
            print(f'file {f} skipped')
            count['skipped'] += 1
            continue

        try:
            source = Path(f).read_text()
        except Exception:
            print(f)
            raise

        try:
            with warnings.catch_warnings(record=True) as log:
                dest = convert(source)
        except NotImplementedError as e:
            print(f'file {f} not supported ({e})')
            count['fatal'] += 1
        else:
            transmogripy_warnings = []
            for w in map(lambda x: x.message, log):
                if isinstance(w, FatalTransmogripyWarning):
                    display = True
                    transmogripy_warnings.append(w)
                elif isinstance(w, TransmogripyWarning):
                    transmogripy_warnings.append(w)
                else:
                    warnings.warn(w)

            if display:
                print(f)
                print(source)
                print('\n ||\n\\||/\n \\/\n')
                print(dest + '\n\n----------\n')
                for w in transmogripy_warnings:
                    print(w)
                break
            elif transmogripy_warnings:
                count['warnings'] += 1
                print(f)
                for w in transmogripy_warnings:
                    print(f'\t{w}')
            else:
                count['ok'] += 1

            base_name = f[len(root_path):-len('.pas')] + '.py'
            dst_name = os.path.join(dst_root, base_name)
            dst_dir = os.path.dirname(dst_name)
            os.makedirs(dst_dir, exist_ok=True)
            Path(dst_name).write_text(dest)

    print(f'\ntotal: {sum(count.values())} files processed')
    for k, v in sorted(count.items(), key=lambda x: x[1], reverse=True):
        if v == 0:
            break  # since the values are sorted, a zero means all the rest are zero too
        print(f'\t{k}: {v} files')



from typing import Iterable

from typing import NamedTuple

from .rule import Rule, ReReplaceRule, NotSupportedRule, ReReplaceFinalRule, HaltRule, EarlyReturnRule
from .segment import PreWord, PostWord, Segment


class RuleSet(NamedTuple):
    pre_words: Segment
    post_words: Segment
    rules: Iterable[Rule]


def get_rules(result_as_var: bool, allow_numpy=True, disclose=True, pre_raw_parts=(), post_raw_parts=()) -> RuleSet:
    pre_words = PreWord()
    post_words = PostWord()

    if disclose:
        pre_words.add_part('disclose')
    for prp in pre_raw_parts:
        pre_words.add_raw(prp)
    for prp in post_raw_parts:
        post_words.add_raw(prp)

    # remember: all these patterns are compiled with the IGNORECASE flag
    rules = (
        # comment rules
        # all these comment rules have 2 modes: one for whole-line comment, and one for end-of-line comment
        # (where it adds 2 spaces)
        ReReplaceFinalRule(r'^(?P<indent>\s*)\{\s*(?P<com>([^$].*)?)\s*\}\s*', '\g<indent># \g<com>'),

        ReReplaceFinalRule('\s*\{\s*(?P<com>([^$].*)?)\s*\}\s*', '  # \g<com>'),

        ReReplaceFinalRule(r'^(?P<indent>\s*)\(\*\s*(?P<com>([^$].*)?)\s*\*\)\s*', '\g<indent># \g<com>'),

        ReReplaceFinalRule('\s*\(\*\s*(?P<com>([^$].*)?)\s*\*\)\s*', '  # \g<com>'),

        ReReplaceFinalRule(r'^(?P<indent>\s*)//\s*(?P<com>.*)', '\g<indent># \g<com>'),

        ReReplaceFinalRule(r'\s*//\s*(?P<com>.*)', '  # \g<com>'),

        # raw string

        ReReplaceFinalRule("'(?P<inner>[^']*)'", "'\g<inner>'"),

        # for loop

        ReReplaceRule('for\s+(?P<var_name>[^ ]+)\s*:=\s*(?P<start>.*)\s+to\s+(?P<end>.*)\s+do',
                      'for \g<var_name> in range(\g<start>, \g<end>+1):'),

        # remove semicolons

        ReReplaceRule(';', ''),

        # begin marks function start

        ReReplaceRule('^begin$', 'def main():'),

        # result rules

        ReReplaceRule.maybe(result_as_var)('(?<![_a-z0-9])Result(?![_0-9a-z])', '__Return__'),

        EarlyReturnRule.maybe(not result_as_var)('(?<=[a-z0-9_])[^a-z0-9_]+Result(?![_0-9a-z])'),
        ReReplaceRule.maybe(not result_as_var)('^(?P<indent>\s*)Result\s*:=\s*', '\g<indent>return '),

        # \\ connector at end of line for line continuation

        ReReplaceRule('(?<![_a-z0-9])(?P<connector>and|or|\+|-|%|^|\||&|\*|/|//)\s*$', r'\g<connector> \\',
                      demand_last_component=True),

        # if result is var, return it at the end

        ReReplaceRule.maybe(result_as_var)('^end\s*\.?\s*$', 'return __Return__', add_prev_indent=True),

        # delete all begins and ends (we trust the source is properly indented)

        ReReplaceRule('(?<![_a-z0-9])begin|end\.?(?![_0-9a-z])\s*', ''),

        # conditional clauses (while/if/elif)

        ReReplaceRule(r'else\s+if\s+(?P<condition>.+)(\s*|\))\sthen',
                      'if \g<condition>:'),

        ReReplaceRule(r'if\s+(?P<condition>.+)(\s*|\))\sthen',
                      'if \g<condition>:'),

        ReReplaceRule(r'while\s+(?P<condition>.+)(\s*|\))\sdo',
                      'while \g<condition>:'),

        # repeat/ until

        ReReplaceRule('(?<![_a-z0-9])repeat(?![_0-9a-z])', 'while True:'),

        ReReplaceRule('^\s*(?<![_a-z0-9])until\s+(?P<condition>.+)', 'if \g<condition>: break', add_prev_indent=True),

        # then

        ReReplaceRule('(?<![_a-z0-9])then(?![_0-9a-z])$', ':'),

        # else

        ReReplaceRule('(?<![_a-z0-9])else(?![_0-9a-z])', 'else:'),

        # all *= rules

        ReReplaceRule('(?<![<>:!])=', '=='),

        ReReplaceRule('(?P<var_name>[^\s]+)\s*:=\s*', '\g<var_name> ='),

        ReReplaceRule('<>', '!='),

        # functions/operators

        ReReplaceRule('(?<![_a-z0-9])length', 'len'),

        ReReplaceRule.maybe(allow_numpy)('(?<![_a-z0-9])length2\(', 'np.size(', hook=pre_words.partial_part('numpy')),

        ReReplaceRule('(?<![_a-z0-9])(str|float|int)to(?P<dest>str|float|int)\s*\(', '\g<dest>(', conv=str.lower),

        ReReplaceRule('(?<![_a-z0-9])exp\s*\(', 'math.exp(', hook=pre_words.partial_part('math')),

        ReReplaceRule('(?<![_a-z0-9])(ln|log)(?P<base>[0-9]*)\s*\(', 'math.log\g<base>(',
                      hook=pre_words.partial_part('math')),

        ReReplaceRule('(?<![_a-z0-9])floor\s*\(', 'math.floor(',
                      hook=pre_words.partial_part('math')),

        ReReplaceRule('(?<![_a-z0-9])ceil\s*\(', 'math.ceil(',
                      hook=pre_words.partial_part('math')),

        ReReplaceRule('(?<![_a-z0-9])power\s*\(', 'pow('),

        ReReplaceRule('(?<![_a-z0-9])random\s*\(', 'random.uniform(0,', hook=pre_words.partial_part('random')),

        ReReplaceRule('(?<![_a-z0-9])randomint\s*\((?P<args>.*)\)', 'random.randint(0,\g<args>-1)',
                      hook=pre_words.partial_part('random')),

        ReReplaceRule('(?<![_a-z0-9])normaldistribution\s*\(', 'random.normalvariate(',
                      hook=pre_words.partial_part('random')),

        ReReplaceRule.maybe(allow_numpy)('(?<![_a-z0-9])SetArray[123]\((?P<lengths>.*)\)', 'np.zeros((\g<lengths>))',
                                         hook=pre_words.partial_part('numpy')),

        ReReplaceRule('(?<![_a-z0-9])SetArray\((?P<length>.*)\)',
                      '[None]*\g<length>'),

        ReReplaceRule('(?<![_a-z0-9])nil(?![_0-9a-z])',
                      'None'),

        ReReplaceRule('\$(?P<num>[a-f0-9]+)',
                      '0x\g<num>'),

        ReReplaceRule('\s+div\s+',
                      '//'),

        # all new rules go BEFORE the NotSupportedRules

        NotSupportedRule('(\{|#|\(\*)\s*\$', 'pre-processor directives not supported'),

        NotSupportedRule('(?<![_a-z0-9])goto\s', 'goto statements are not supported'),

        # this rule goes absolutely last

        HaltRule(),
    )

    return RuleSet(pre_words=pre_words, post_words=post_words, rules=rules)



from typing import Dict, ClassVar

from .__data__ import __version__

from functools import partial


class Segment:
    """
    A collection of parts that appears in the output file.
    This class is meant to be subclassed with a PARTS class member.
    """

    PARTS: ClassVar[Dict['str', 'str']]  # virtual member

    def __init__(self):
        self.raw_parts = []
        self.activated = set()  # so we don't add the same part twice

    def add_part(self, part_name):
        if part_name in self.activated:
            return
        val = self.PARTS[part_name]
        self.raw_parts.append(val)
        self.activated.add(part_name)

    def add_raw(self, part_string):
        self.raw_parts.append(part_string)

    def partial_part(self, part_name):
        return partial(self.add_part, part_name)

    def __iter__(self):
        yield from self.raw_parts

    def join(self, sep='\n'):
        return sep.join(self)


class PreWord(Segment):
    PARTS = {
        'numpy': 'import numpy as np',
        'math': 'import math',
        'random': 'import random',
        'disclose':
            f'"""\nthis script was automatically converted from pascal using transmogripy v{__version__}\n"""\n',
    }

    def join(self, *args, **kwargs):
        return super().join(*args, **kwargs) + '\n' * 2


class PostWord(Segment):
    PARTS = {}

    def join(self, *args, **kwargs):
        return '\n' * 2 + super().join(*args, **kwargs)



import ast


def is_valid_python(code):
    try:
        ast.parse(code)
    except SyntaxError as e:
        return False, e
    return True, None


def isblank(s: str):
    """
    >>> assert isblank('')
    >>> assert isblank('  ')
    >>> assert not isblank('   a    ')
    """
    return not s or s.isspace()


class TransmogripyWarning(UserWarning):
    """
    A warning indicating the resulting script might not work as intended
    """
    pass


class FatalTransmogripyWarning(TransmogripyWarning):
    """
    A warning indicating the result script doesn't pass basic sanity checks
    """
    pass


class EarlyReturnDetected(Exception):
    """
    An exception indicating an early return
    """


__all__ = ['is_valid_python', 'isblank', 'TransmogripyWarning', 'FatalTransmogripyWarning', 'EarlyReturnDetected']



from typing import Union, Callable, Optional, List

from abc import ABC, abstractmethod

import re

from .__util import *


class Final(str):
    """
    a string component in a line that should not be run through any more rules
    """
    __slots__ = ()
    pass


class LineComponents:
    """
    Like a list, but setting a cell with an iterable instead inserts it at the area
    >>> x = LineComponents('')
    >>> x.parts = ['a', 'bcd', 'e']
    >>> ' '.join(x)
    'a bcd e'
    >>> x[1] = ['b', 'c', 'd']
    >>> ' '.join(x)
    'a b c d e'
    """

    def __init__(self, start: str):
        self.parts: List[str] = [start]

    def non_final_part(self):
        """
        get the index and value of the first element in the list that is not a member of the Final class
        """
        return next(
            ((i, p) for (i, p) in enumerate(self.parts) if not isinstance(p, Final)),
            None)

    def __setitem__(self, key, value):
        if not isinstance(value, str):
            key = slice(key, key + 1)
            value = (v for v in value if v)
        self.parts[key] = value

    def __iter__(self):
        yield from self.parts

    def __len__(self):
        return len(self.parts)


class Rule(ABC):
    @abstractmethod
    def __call__(self, line, env) -> Union[str, None, List[Union[str, Final]]]:
        """
        :return: None if the line hasn't changed, or the line if it is changed,
            a list if the rule causes a split in the captured pattern.
        """
        return None

    @classmethod
    def maybe(cls, cond):
        if cond:
            return cls
        return _NilRule


class _NilRule(Rule):
    _single = None

    def __new__(cls, *args, **kwargs):
        if not cls._single:
            cls._single = super().__new__(cls)
        return cls._single

    def __call__(self, line, env):
        return None


class PatternRule(Rule, ABC):
    def __init__(self, pattern: str):
        self.pattern = re.compile(pattern, re.IGNORECASE)


class EarlyReturnRule(PatternRule):
    def __call__(self, line, env):
        if self.pattern.search(line):
            raise EarlyReturnDetected()
        return None


class ReReplaceRule(PatternRule):
    def __init__(self, pattern: str, sub: str, add_prev_indent=False, hook: Optional[Callable[[], None]] = None,
                 demand_last_component=False, conv: Optional[Callable[[str], str]] = None):
        super().__init__(pattern)
        self.sub = sub
        self.add_prev_indent = add_prev_indent
        self.hook = hook
        self.demand_last_component = demand_last_component
        self.output_convert = conv

    def __call__(self, line, env):
        if self.demand_last_component and not env['last_component']:
            return None
        ret, n = self.pattern.subn(self.sub, line)
        if n == 0:
            return None
        if self.output_convert:
            ret = self.output_convert(ret)
        if self.add_prev_indent:
            ret = env['prev_indent'] + ret
        if self.hook:
            self.hook()
        return ret


class ReReplaceFinalRule(PatternRule):
    def __init__(self, pattern: str, sub: str):
        super().__init__(pattern)
        self.sub = sub

    def __call__(self, line, env):
        ret = []
        while True:
            next_match = self.pattern.search(line)
            if not next_match:
                if not ret:
                    return None
                ret.append(line)
                break
            ret.append(line[:next_match.start()])
            match_sub = next_match.expand(self.sub)
            ret.append(Final(match_sub))
            line = line[next_match.end():]
        return ret


class NotSupportedRule(PatternRule):
    def __init__(self, pattern: str, msg: str):
        super().__init__(pattern)
        self.msg = msg

    def __call__(self, line, env):
        if self.pattern.search(line):
            raise NotImplementedError(self.msg)
        return None


class HaltRule(Rule):
    def __call__(self, line, env):
        return [Final(line)]



from typing import Iterable, Union, List, Tuple, Match

from enum import Enum
from collections import namedtuple
import re

from .__util import *


class ScanMode(Enum):
    code = 'code', None, None
    raw = 'raw', "'", "'"
    bracket_comment = 'com', '{', '}'
    star_comment = 'com', '(*', '*)'
    line_comment = 'com', '//', ''


class Comment(str):
    __slots__ = ('args',)

    @classmethod
    def full(cls, arg, open_, close):
        return open_ + arg.strip() + close

    def __new__(cls, arg: str, open_, close):
        full = cls.full(arg, open_, close)
        ret = super().__new__(cls, full)
        ret.args = arg, open_, close
        return ret

    def __repr__(self):
        return type(self).__name__ + '(' + ', '.join(self.args) + ')'


class LineParts(List[str]):
    __slots__ = ('remove_inline_comments',)

    def __new__(cls, *args, remove_inline_comments=False, **kwargs):
        return super().__new__(cls, *args, **kwargs)

    def __init__(self, *args, remove_inline_comments=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.remove_inline_comments = remove_inline_comments

    @property
    def last_is_comment(self):
        for i in reversed(self):
            if isinstance(i, Comment):
                return True
            if isblank(i):
                continue
            return False
        return False

    def append(self, o, mode=None):
        if mode and mode.value[0] == 'com':
            o = Comment(o, *mode.value[1:])
        if not isinstance(o, Comment) and not o.isspace():
            while self.last_is_comment:
                if self.remove_inline_comments:
                    self.pop()
                    continue
                raise NotImplementedError('inline comments must be at the end of a line')
        super().append(o)

    def join(self):
        return ''.join(str(i) for i in self)


Region = namedtuple('Region', 'start_pattern end_pattern mode')
regions = [
    Region(re.compile("'"), re.compile(r"(?<!\\)'"), ScanMode.raw),
    Region(re.compile('{(?!\$)'), re.compile('}'), ScanMode.bracket_comment),
    Region(re.compile('\(\*'), re.compile('\*\)'), ScanMode.star_comment),
    Region(re.compile('//'), re.compile('$'), ScanMode.line_comment),
]


def next_region_ind(line):
    """
    returns the index of the first matching region. If not found, returns the line's length
    >>> next_region_ind("abc{d'ef")
    3
    >>> next_region_ind("re'ge{x")
    2
    >>> next_region_ind("regex")
    5
    """
    matches = [reg.start_pattern.search(line) for reg in regions]
    match = min(matches,
                key=lambda m: float('inf') if m is None else m.start())
    if match is None:
        return len(line)

    return match.start()


def match_region(line) -> Tuple[Region, Match]:
    """
    returns the region that matches the first characters in the line, raises an error if there is more or less than one
    matching region. Also returns the match
    """
    matches = [(reg, reg.start_pattern.match(line)) for reg in regions]
    matches = [(k, v) for (k, v) in matches if v]
    if len(matches) != 1:
        raise AssertionError(f'expected 1 match, got {len(matches)}')
    return matches[0]


def filter_multiline_comments(source: Iterable[str], remove_inline_comments=False):
    """
    all this function does is wrap all lines in between multiline comments with a seperate multiline comment
        {a
        b
        c}
    to:
        {a}
        {b}
        {c}
    >>> assert list(filter_multiline_comments(["abcd","efgh"])) == ["abcd","efgh"]
    >>> assert list(filter_multiline_comments(["abc{def","def","def}"])) == ["abc{def}","{def}","{def}"]
    >>> assert list(filter_multiline_comments(["abc{def","   def","def}"])) == ["abc{def}","   {def}","{def}"]
    >>> assert list(filter_multiline_comments(["my mother said 'i need {some} dire frui{{t'" \
                                                ,"{she didn't actually","say","that}" \
                                                ,"but you {don't care} {do you?}"])) \
                                                 == ["my mother said 'i need {some} dire frui{{t'" \
                                                 ,"{she didn't actually}","{say}","{that}" \
                                                 ,"but you {don't care} {do you?}"]
    >>> assert list(filter_multiline_comments(['hi //there','how are you?','{i','am}','//fi{ne','thanks', \
                                                '{no //problem}'])) \
                                                == ['hi //there', 'how are you?', '{i}', '{am}', '//fi{ne', 'thanks',\
                                                 '{no //problem}']
    """

    region: Union[Region, type(...), None] = None
    # the region we are in, None means we are in code
    # ... means the next token is the start of a region, and we need to find out which
    for line in source:
        line_parts = LineParts(remove_inline_comments=remove_inline_comments)
        if region is not None:
            indent = re.match('\s*', line)
            line_parts.append(indent.group())
            line = line[indent.end():]
        capture = ''
        while line:
            if region is None:
                code_scan_length = next_region_ind(line)
                if code_scan_length == 0:
                    region = ...
                else:
                    line_parts.append(line[:code_scan_length])
                    line = line[code_scan_length:]
            elif region is ...:
                region, match = match_region(line)
                if region.mode.value[0] == 'com':
                    capture = ''
                elif region.mode == ScanMode.raw:
                    capture = match[0]
                else:
                    raise AssertionError('unhandled region mode: ' + str(region.mode))
                line = line[match.end():]
            else:
                # regular region (string or comment)
                end_token_match = region.end_pattern.search(line)
                mode = region.mode
                if not end_token_match:
                    # the rest of the line is the current region (it continues unto the next line but we split it here)
                    # technically, this is only allowed if we're in a comment, but I'm not about to start throwing
                    # pascal errors
                    capture += line
                    line = ''
                else:
                    if region.mode.value[0] == 'com':
                        end_cap_ind = end_token_match.start()
                    elif region.mode == ScanMode.raw:
                        end_cap_ind = end_token_match.end()
                    else:
                        raise AssertionError('unhandled region mode: ' + str(region.mode))
                    capture += line[:end_cap_ind]
                    line = line[end_token_match.end():]
                    region = None
                # either way, this region is done
                line_parts.append(capture, mode)
                capture = ''
        yield line_parts.join()



import setuptools

import transmogripy

setuptools.setup(
    name=transmogripy.__name__,
    version=transmogripy.__version__,
    author=transmogripy.__author__,
    description='tool to convert short pascal scripts to python',
    license='gpl-2.0',
    url='https://github.com/talos-gis/transmogripy',
    packages=['transmogripy'],
    python_requires='>=3.6.0',
    include_package_data=True,
    data_files=[
        ('', ['README.md', 'CHANGELOG.md', 'LICENSE']),
    ],
)



from typing import Union

from enum import Enum
from io import StringIO
import re
import warnings

from .rule import LineComponents, Final
from .rules import get_rules
from .filter_multiline_comments import filter_multiline_comments
from .__util import *


class ResultBehaviour(Enum):
    try_ = 'try'
    var = 'variable'


def convert(pascal: str, check_syntax=True, result_behaviour: Union[str, ResultBehaviour] = 'try', disclose=True,
            remove_inline_comments=True, pre_parts=('from talos import *',), post_parts=()):
    """
    convert a pascal script to a python script
    :param pascal: the pascal script as a single string
    :param check_syntax: whether to check the python syntax of the output and issue a warning if any errors are found
    :param result_behaviour: how to handle the result variable
    :param disclose: whether to add a disclosure tot he output stating it was automatically generated
    :param remove_inline_comments: whether to remove inline comments in the source code. Setting this to false will
        raise an error if inline comments are found.
    :param pre_parts: any text to add before the output code should be entered here
    :param post_parts: any text to add after the output code should be entered here
    :return: the python script as a string
    """
    result_behaviour = ResultBehaviour(result_behaviour)
    result_as_var = result_behaviour == ResultBehaviour.var
    pre_words, post_words, rules = get_rules(result_as_var, disclose=disclose,
                                             pre_raw_parts=pre_parts, post_raw_parts=post_parts)
    env = {}

    lines = list(
        filter_multiline_comments(pascal.splitlines(keepends=False), remove_inline_comments=remove_inline_comments))
    # just treat single-line scripts as though they have a begin and end around them
    if len(lines) in (0, 1):
        lines = ['begin', *('\t' + l for l in lines), 'end']
    else:
        try:
            var_index = lines.index('var')
        except ValueError:
            # if there is no var, then delete nothing (we want to keep anything before begin in case it's a comment)
            pass
        else:
            begin_index = lines.index('begin', var_index)
            del lines[var_index:begin_index]

    ret = StringIO()

    for line in lines:
        # note: we want to pass blank lines only if it was blank in the original!
        if isblank(line):
            ret.write('\n')
            continue

        indent = re.match('^\s*', line).group(0)
        comps = LineComponents(line)
        non_final = comps.non_final_part()
        while non_final:
            comp_ind, comp = non_final
            if not comp:
                comps[comp_ind] = Final(comp)
            else:
                env['last_component'] = comp_ind + 1 == len(comps)
                for rule in rules:
                    try:
                        res = rule(comp, env)
                    except EarlyReturnDetected:
                        return convert(pascal, check_syntax, 'variable', disclose, remove_inline_comments, pre_parts,
                                       post_parts)
                    if res is not None:
                        comp = comps[comp_ind] = res
                        if not isinstance(res, str):
                            break
                        if not comp:
                            comps[comp_ind] = Final(comp)
                            break

            non_final = comps.non_final_part()

        line = ''.join(comps)
        env['prev_indent'] = indent
        if not isblank(line):
            ret.write(line + '\n')

    ret = pre_words.join() + ret.getvalue() + post_words.join()
    ret = ret.rstrip() + '\n'  # add a single trailing newline as per PEP8
    if check_syntax:
        valid, error = is_valid_python(ret)
        if not valid:
            warnings.warn(
                f'the python script did not pass syntax checking, the error was: {error}',
                category=FatalTransmogripyWarning)
    return ret
